//-------------------------------------------------------------------------------------------------+
// AislGlobalPlanningSVC_impl.cpp
// Service implementation code of AislGlobalPlanning.idl
//-------------------------------------------------------------------------------------------------+

#include "AislGlobalPlanningSVC_impl.h"
#include <mmsystem.h>


//-------------------------------------------------------------------------------------------------+
// RXgN^
//-------------------------------------------------------------------------------------------------+
AreaInfoServiceSVC_impl::AreaInfoServiceSVC_impl( SubAreas *subareas, CRITICAL_SECTION *cs ){

	this->subareas = subareas;
	c_section      = cs;
}


//-------------------------------------------------------------------------------------------------+
// fXgN^
//-------------------------------------------------------------------------------------------------+
AreaInfoServiceSVC_impl::~AreaInfoServiceSVC_impl(){
}


//-------------------------------------------------------------------------------------------------+
// TuGAԂ
//-------------------------------------------------------------------------------------------------+
TUT::TimedAreaInfo *AreaInfoServiceSVC_impl::getAreaInfo(){

	EnterCriticalSection( c_section );

	TUT::TimedAreaInfo *area_info = new TUT::TimedAreaInfo();

	area_info->tm.sec  = timeGetTime() / 1000;
	area_info->tm.nsec = ( timeGetTime() % 1000 ) * 1000000;

	area_info->data.length( subareas->size() );

	for( int i=0; i<(int)subareas->size(); i++ ){

		area_info->data[i].id         = (*subareas)[i].ID();
		area_info->data[i].population = (*subareas)[i].population;

		Vertexes vertexes = (*subareas)[i].Vertex();

		area_info->data[i].vertexes.length( vertexes.size() );

		for( int j=0; j<(int)vertexes.size(); j++ ){

			area_info->data[i].vertexes[j].x = vertexes[j].x;
			area_info->data[i].vertexes[j].y = vertexes[j].y;
		}

		vector<int> ids = (*subareas)[i].ConnectedAreaIDs();

		area_info->data[i].links.length( ids.size() );

		for( int j=0; j<(int)ids.size(); j++ ){

			area_info->data[i].links[j].id = ids[j];

			Segment seg = (*subareas)[i].Border( ids[j] );

			area_info->data[i].links[j].boder.p1.x = seg.x;
			area_info->data[i].links[j].boder.p1.y = seg.y;
			area_info->data[i].links[j].boder.p2.x = seg.x + seg.dx;
			area_info->data[i].links[j].boder.p2.y = seg.y + seg.dy;
		}
	}

	LeaveCriticalSection( c_section );

	return area_info;
}

