//------------------------------------------------------------------------------------------------+
// * @file  EnvironmentSimulator.h
// * @brief Simulate robot and people movement and provide connecting RTC sensor information
//------------------------------------------------------------------------------------------------+

#ifndef ENVIRONMENTSIMULATOR_H
#define ENVIRONMENTSIMULATOR_H

// RT֘Awb_
#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>
#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

// Service implementation headers
#include "MRFCSVC_impl.h"
#include "AislGlobalPlanningSVC_impl.h"


// Service Consumer stub headers
#include "IIS2Stub.h"
#include "LaserRangeSensorDataStub.h"

// dN\wb_t@CCN[h𓾂Ȃ
#include "Simulator/Simulator.h"
#include "Simulator/Window.h"
#include "RangeSensor.h"

// ΐlnɂȂꂽwb_t@CCN[hĒ܂
#include "H/OdometryCalcClass.h"
#include "H/TIMER_CLASS.h"
#include "PathPlanner/base_data_light.h"


//------------------------------------------------------------------------------------------------+
// }N
//------------------------------------------------------------------------------------------------+
#define SIMULATION_INTERVAL		0.1			// ̃V~[VŌo߂鎞[b]


using namespace RTC;
using namespace IIS;
using namespace MRFC;
using namespace SensorRTC::LaserRangeSensor::idl;


//------------------------------------------------------------------------------------------------+
// EnvironmentSimulator
//------------------------------------------------------------------------------------------------+
class EnvironmentSimulator : public RTC::DataFlowComponentBase{

public:
	EnvironmentSimulator(RTC::Manager* manager);
	~EnvironmentSimulator();

	// eANeBreB
	virtual RTC::ReturnCode_t onInitialize();
	virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);
	virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);
	virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

protected:

	// RtBO[V֌W̕ϐ
	std::string m_FieldDataFile;
	double m_robot_tread;
	double m_robot_acceleration;
	double m_robot_deceleration;
	double m_robot_init_x;
	double m_robot_init_y;
	double m_robot_init_heading;
	double m_robot_move_error_x;
	double m_robot_move_error_y;
	double m_robot_move_error_heading;
	double m_walkingrate_mean;
	double m_walkingrate_std_deviation;
	int    m_localmap_width;
	int    m_localmap_height;
	double m_localmap_scale;
	double m_globalmap_scale;
	double m_show_map_scale;
	int    m_following_target_id;
	double m_range_start_position;
	double m_range_end_position;
	double m_range_max_distance;
	long   m_range_far_code;
	int    m_range_data_num;
	double m_range_sensor_x;
	double m_range_sensor_y;
	double m_range_sensor_heading;
	double m_range_std_deviation;
	double m_range_person_radius;
	int    m_display_range_data;
	int    m_simulate_people;
	int    m_simulate_skip;

	// DataInPort declaration
	IIS::TimedVelocity2D m_robotControlInfo;
	InPort<IIS::TimedVelocity2D> m_robotControlInfoIn;

	// DataOutPort declaration
	TimedMeasuredData m_rangedata;
	OutPort<TimedMeasuredData> m_rangedataOut;
	IIS::TimedPose2D m_robot_state;
	OutPort<IIS::TimedPose2D> m_robot_stateOut;
	IIS::TimedVelocity2D m_robot_velocity;
	OutPort<IIS::TimedVelocity2D> m_robot_velocityOut;
	TUT::TimedPersonsGlobalPosition m_persons;
	OutPort<TUT::TimedPersonsGlobalPosition> m_personsOut;
	TimedPeopleTrackingData m_peopledata;
	OutPort<TimedPeopleTrackingData> m_peopledataOut;

	// CORBA Port declaration
	RTC::CorbaPort m_AbsoluteMapServicePort;
	RTC::CorbaPort m_RelativeMapServicePort;
	RTC::CorbaPort m_AreaInfoServicePort;

	// Service declaration
	AbsoluteMapServiceSVC_impl m_AbsoluteMapService;
	RelativeMapServiceSVC_impl m_RelativeMapService;
	AreaInfoServiceSVC_impl m_AreaInfoService;

private:
	void ExecuteSimulation();
	void OutputRobotState();
	void OutputRobotVelocity();
	void OutputPersons();
	void OutputPeopleData();
	void OutputRangeData();
	void MakeGlobalMap();
	void MakeLocalMap();
	void UpdateRobotInfo();
	void DrawWindow();

	ROBOT::ODOMETRY odometry;
	TIMER::TIMER_CLASS *cTimer;

	RTC::Time time;

	// For Service ports
	CRITICAL_SECTION cs_globalmap;
	CRITICAL_SECTION cs_localmap;
	CRITICAL_SECTION cs_subareas;
	TimedAbsoluteOGMapData globalmap;
	TimedRelativeOGMapData localmap;

	// For Simulator
	Simulator *sim;
	Field field;
	ActionMng actions;
	SubAreas subareas;
	Robot robot;
	RoboMap map;
	int step;
	SimulationWindow *window;
	RangeSensor sensor;

};



//------------------------------------------------------------------------------------------------+
extern "C"
{
  DLL_EXPORT void EnvironmentSimulatorInit(RTC::Manager* manager);
};

#endif // ENVIRONMENTSIMULATOR_H
