//------------------------------------------------------------------------------------------------+
// Filename: RangeSensor.cpp
// Content : ZTV~[^ōČ邽߂̃NX̊֐`
//------------------------------------------------------------------------------------------------+

#include "RangeSensor.h"
#include "Simulator/applib.h"
#include <cmath>


//------------------------------------------------------------------------------------------------+
// RXgN^
//------------------------------------------------------------------------------------------------+
RangeSensor::RangeSensor( double start_pos, double end_pos, double max_dist, long far_code, int data_num, double x, double y, double heading, double std_dev ){

	start_position = start_pos;
	end_position   = end_pos;
	max_distance   = max_dist;
	this->far_code = far_code;
	this->data_num = data_num;
	sensor_x       = x;
	sensor_y       = y;
	sensor_heading = heading;
	this->std_dev  = std_dev;

	if( end_position < start_position ){

		start_position = end_pos;
		end_position   = start_pos;
	}

	if( data_num < 0 ) data_num = 0;

	range_data.resize( data_num );
	rays.resize( data_num );
}


//------------------------------------------------------------------------------------------------+
// f[^擾\Ԃ
//------------------------------------------------------------------------------------------------+
double RangeSensor::DataInterval(){

	if( data_num <= 0 ) return 0;

	return ( end_position - start_position ) / data_num;
}


//------------------------------------------------------------------------------------------------+
// ZTɂǂƐl̔Fs
//------------------------------------------------------------------------------------------------+
void RangeSensor::Sensing( Robot *robo, Walls *walls, People *people, double peason_radius ){

	double sensor_pos_x = robo->x + sensor_x * cos( robo->theta ) - sensor_y * sin( robo->theta );
	double sensor_pos_y = robo->y + sensor_x * sin( robo->theta ) + sensor_y * cos( robo->theta );

	double direction = robo->theta + sensor_heading + ( start_position - 90.0 ) * deg_to_rad;

	for( int i=0; i<data_num; i++ ){

		// Vectorpӂivւ̒PʃxNg߂j
		Line ray( sensor_pos_x, sensor_pos_y, cos( direction ), sin( direction ));
		rays[i] = ray;

		ray *= max_distance;

		range_data[i] = far_code;

		// 蔻s
		Sensing( i, ray, walls );
		if( 0.0 < peason_radius ) Sensing( i, ray, people , peason_radius );

		if( range_data[i] != far_code ){

			rays[i] *= (double)range_data[i] / 1000.0;
			range_data[i] += (long)( 1000.0 * GausRand( 0.0, std_dev ).Rand() );
		}
		else rays[i] *= max_distance;

		// pxXV
		direction += DataInterval() * deg_to_rad;
	}
}


//------------------------------------------------------------------------------------------------+
// ZŤ1{ɂǂ̔Fs
// range_data[no]Ɍʂ
//------------------------------------------------------------------------------------------------+
void RangeSensor::Sensing( int no, Line ray, Walls *walls ){

	// 蔻s
	for( int w=0; w<walls->WallNum(); w++ ){

		double t = 1000.0 * rays[no].Get_t( walls->Wall(w) );

		if( !ray.KreuzungUrteilX( walls->Wall(w) ) ) continue;
		if( range_data[no] != far_code && range_data[no] <= t ) continue;

		range_data[no] = (long) t;
	}
}


//------------------------------------------------------------------------------------------------+
// ZŤ1{ɂl̔Fs
// range_data[no]Ɍʂ
//------------------------------------------------------------------------------------------------+
void RangeSensor::Sensing( int no, Line ray, People *people, double peason_radius ){

	for( int p=0; p<(int)people->size(); p++ ){

		Line p_to_ray = ray.PointToLine( (*people)[p].p );

		if( peason_radius < p_to_ray.Length() ) continue;

		double t = 1000.0 * ( rays[no].Get_t( p_to_ray ) - peason_radius + p_to_ray.Length() );

		if( range_data[no] != far_code && range_data[no] <= t ) continue;

		range_data[no] = (long) t;
	}
}

