//------------------------------------------------------------------------------------------------+
// Filename: Simulator.cpp
// Content : V~[^NX̊֐`
//------------------------------------------------------------------------------------------------+

#include "Simulator.h"
#include <cmath>
#include <iostream>


//------------------------------------------------------------------------------------------------+
// }N̒`
//------------------------------------------------------------------------------------------------+
#define PERSON_RADIUS		0.225	// l̔a[m]
#define VIEW_RANGE			7.5		// l̎̍ő勗[m]
#define PS_RATIO			1.2		// p[\iXy[X
#define Q_PERSON			0.6		// l̃|eV肷萔
#define Q_ROBOT				0.8		// {bg̃|eV肷萔
#define Q_WALLS				0.05	// ǂ̃|eV肷萔
#define SEAT_RADIUS			0.45	// Ȃ̓B蔼a
#define QUEUE_RADIUS		0.3		// s̐擪B蔼a
#define FRUIN_CONST			0.341	// Qlɂ萔
#define BORDER_SIDE			0.5		// oԖړIn̒[̕␳
#define WALL_DIST_THRESHOLD	0.8		// 臒lǂ΃|eV̓[ 0ȉ̒lw肷臒lƔf


// Name: Simulator() -----------------------------------------------------------------------------+
// RXgN^
//------------------------------------------------------------------------------------------------+
Simulator::Simulator( Field *field, ActionMng *action_mng, SubAreas subareas, double walkingrate, double walkingrate_dev ){

	person_id   = 0;
	this->field = field;
	actions     = action_mng;
	network     = Network( subareas );

	mean_speed  = walkingrate;
	sigma_speed = walkingrate_dev;

	seats_mng.resize( field->seats.DataNum() );
	queue_mng.resize( field->queues.DataNum() );
}


// Name: Simulation() ----------------------------------------------------------------------------+
// pass_timebԂ̐lX̓Č
//------------------------------------------------------------------------------------------------+
void Simulator::Simulation( double pass_time ){

	try{

		// ͈͊O̐lł
		EraseOutside();

		// l̈ړs
		people = Moving( pass_time );

		// l̓ꏈ
		PeopleEnter( pass_time );

		// l̈ʒu\
		People predict = Moving( pass_time );

		// ړlbg[N̍Gf[^XV
		network.ResetPopulation();
		for( int i=0; i< (int)predict.size(); i++ ) network.AddPopulation( predict[i].p );

		// ǗĂړIn̐lXV
		CheckSeatsFree();

		// l̍s
		for( int i=0; i<(int)people.size(); i++ ){

			// ҋ@ƍs̍XV
			if( people[i].wait_time ){

				people[i].wait_time -= pass_time;
				people[i].p.dx = people[i].p.dy = 0.0;

				if( people[i].wait_time <= 0.0 ){

					people[i].wait_time = 0.0;
					if( people[i].action.Type() == GOTO_QUEUE ) OutQueue( people[i] );
					people[i].action = actions->NextAction( people[i].action );

					// oHTƖړIn̑I
					people[i] = RoutePlanning( people[i] );
				}
			}
			else{

				// ԖړInB
				if( ReachWaypoint( people[i] ) ) people[i].path.pop_front();

				// ړInBƑҋ@Ԃ̐ݒ܂͍s̍XV
				if( ReachDestination( people[i] ) ){

					people[i].wait_time = people[i].action.WaitTime();

					if( people[i].wait_time == 0.0 ){

						if( people[i].action.Type() == GOTO_QUEUE ) OutQueue( people[i] );
						people[i].action = actions->NextAction( people[i].action );

						// oHTƖړIn̑I
						people[i] = RoutePlanning( people[i] );
					}
				}
				// ȂɂƂꂽꍇ̖ړIn̍Đݒ
				else if( people[i].action.Type() == GOTO_SEAT
					   && seats_mng[ people[i].action.dest_no ].sitting
					   && seats_mng[ people[i].action.dest_no ].id != people[i].ID() ){

					people[i] = RoutePlanning( people[i] );
				}
				// ɕłƂ
				else if( people[i].action.Type() == GOTO_QUEUE && ToQueue( people[i] ).Length() < QUEUE_RADIUS ){
					people[i].p.dx = people[i].p.dy = 0.0;
					continue;
				}
/*				// oHOƍČvZ
				else{
					if( !people[i].path.empty() &&
						network.SubAreaNo( people[i].p ) != -1 &&
						!network.GetSubArea( people[i].p ).CheckConnect( people[i].path.front() ) ){ cerr << "Error?\n"; }//people[i] = RoutePlanning( people[i] ); }
				}
*/
				// |eV@ɐlX̑xxNgXV
				people[i] = Potential( people[i], predict );
			}
		}
	}
	catch( const char *e ){
		std::cerr << e << std::endl; 
		throw;
	}
	catch(...){
		std::cerr << "Error: Unknown error." << std::endl; 
		throw;
	}
}


// Name: Moving() --------------------------------------------------------------------------------+
// pass_timeb̐lXf[^Ԃ
//------------------------------------------------------------------------------------------------+
People Simulator::Moving( double pass_time ){

	People after_move_people;

	// ړWvZ
	for( int i=0; i < (int)people.size(); i++ ){

		Person pi = people[i];

		pi.p.x += pass_time * pi.p.dx;
		pi.p.y += pass_time * pi.p.dy;

		after_move_people.push_back( pi );
	}

	return after_move_people;
}


// Name: EraseOutside() --------------------------------------------------------------------------+
// ͈͊O̐lł
//------------------------------------------------------------------------------------------------+
void Simulator::EraseOutside(){

	People survive;

	for( int i=0; i<(int)people.size(); i++ ){

		Line v = people[i].p;

		if( 0.0 <= v.x && 0.0 <= v.y && v.x <= field->FieldWidth() && v.y <= field->FieldLength() ){

			if( !ReachExit( people[i] ) ) survive.push_back( people[i] );
		}
	}
	people = survive;
}


// Name: PeopleEnter() ---------------------------------------------------------------------------+
// l̓
//------------------------------------------------------------------------------------------------+
void Simulator::PeopleEnter( double pass_time ){

	for( int i=0; i<field->EntranceNum(); i++ ){

		if( field->PersonEnter( i, pass_time ) ){

			Person in = Person( person_id++, GausRand( mean_speed, sigma_speed ).Rand() );
			in.p = field->EnterPosition( i );
			in.action = actions->FirstAction( i );
			in = RoutePlanning( in );					// oHTƖړIn̑I
			people.push_back( in );
		}
	}
}


// Name: ReachDestination() ----------------------------------------------------------------------+
// ړInB
//------------------------------------------------------------------------------------------------+
bool Simulator::ReachDestination( Person p ){

//	if( ReachExit(p)  ) return true;
	if( ReachPoint(p) ) return true;
	if( ReachSeat(p)  ) return true;
	if( ReachQueue(p) ) return true;
	return false;
}


// Name: ReachExit() -----------------------------------------------------------------------------+
// ړIniojB
//------------------------------------------------------------------------------------------------+
bool Simulator::ReachExit( Person p ){

	if( p.action.Type() != GOTO_EXIT ) return false;
	p.p *= -1.0;
	if( p.p.KreuzungUrteilX( field->exits.GetData( p.action.dest_no ).GetExit() )) return true;
	return false;
}


// Name: ReachPoint() ----------------------------------------------------------------------------+
// ړInin_jB
//------------------------------------------------------------------------------------------------+
bool Simulator::ReachPoint( Person p ){

	if( p.action.Type() != GOTO_POINT ) return false;

	Point dest = field->points.GetData( p.action.dest_no );

	if( Line( p.p.x, p.p.y, p.p.x - dest.GetX(), p.p.y - dest.GetY() ).Length() <= dest.Radius() ){
		return true;
	}
	return false;
}


// Name: ReachSeat() -----------------------------------------------------------------------------+
// ړIniȁjB
//------------------------------------------------------------------------------------------------+
bool Simulator::ReachSeat( Person p ){

	if( p.action.Type() != GOTO_SEAT ) return false;

	Seat dest = field->seats.GetData( p.action.dest_no );

	if( Vector2D( p.p.x - dest.GetX(), p.p.y - dest.GetY() ).Length() <= SEAT_RADIUS ){
		seats_mng[p.action.dest_no].sitting = true;
		seats_mng[p.action.dest_no].id      = p.ID();
		return true;
	}
	return false;
}


// Name: ReachQueue() ----------------------------------------------------------------------------+
// ړInisjB
//------------------------------------------------------------------------------------------------+
bool Simulator::ReachQueue( Person p ){

	if( p.action.Type() != GOTO_QUEUE || QUEUE_RADIUS < ToQueue( p ).Length() ) return false;

	// ɉĂ邩mF
	for( std::list<int> list = queue_mng[p.action.dest_no]; !list.empty(); list.pop_front() ){
		if( list.front() == p.ID() ){

			// ̐擪ǂmF
			if( queue_mng[p.action.dest_no].front() != p.ID() ) return false;
			return true;
		}
	}

	// ܂ɉĂȂ΍Ōɉ
	queue_mng[p.action.dest_no].push_back( p.ID());
	return false;
}


// Name: ReachWaypoint() -------------------------------------------------------------------------+
// ԖړInB
//------------------------------------------------------------------------------------------------+
bool Simulator::ReachWaypoint( Person p ){

	if( p.path.empty() ) return false;

//	if( p.path.front().InSide( p.p )) return true;

	Path path = p.path;

	while( !path.empty() ){

		if( path.front().InSide( p.p ) ) return true;
		path.pop_front();
	}

	return false;
}


// Name: RoutePlanning() -------------------------------------------------------------------------+
// ړIn̑IƌoHv
//------------------------------------------------------------------------------------------------+
Person Simulator::RoutePlanning( Person p ){

	double best_cost;
	GausRand gaus( 1.0, 0.3 );		// ړInÎ΂Čp̗

	switch( p.action.Type() ){
		case GOTO_EXIT:

			for( int i=0; i< field->exits.DataNum( p.action.DestID() ); i++ ){

				Coordinate dest;
				dest.x = field->exits.GetData( p.action.DestID(), i ).GetX();
				dest.y = field->exits.GetData( p.action.DestID(), i ).GetY();
				PathCost path = network.PathFinding( p.p, dest );
				path.cost    *= gaus.Rand();

				if( i==0 || path.cost < best_cost ){
					best_cost = path.cost;
					p.path = path.path;
					p.action.dest_no = field->exits.GetDataNo( p.action.DestID(), i );
				}
			}
			break;

		case GOTO_POINT:

			for( int i=0; i< field->points.DataNum( p.action.DestID() ); i++ ){

				Coordinate dest;
				dest.x = field->points.GetData( p.action.DestID(), i ).GetX();
				dest.y = field->points.GetData( p.action.DestID(), i ).GetY();
				PathCost path = network.PathFinding( p.p, dest );
				path.cost    *= gaus.Rand();

				if( i==0 || path.cost < best_cost ){
					best_cost = path.cost;
					p.path = path.path;
					p.action.dest_no = field->points.GetDataNo( p.action.DestID(), i );
				}
			}
			break;

		case GOTO_SEAT:

			best_cost = 99999999999999999999999999.9;

			for( int i=0; i< field->seats.DataNum( p.action.DestID() ); i++ ){

				if( seats_mng[ field->seats.GetDataNo( p.action.DestID(), i ) ].sitting ) continue;

				Coordinate dest;
				dest.x = field->seats.GetData( p.action.DestID(), i ).GetX();
				dest.y = field->seats.GetData( p.action.DestID(), i ).GetY();
				PathCost path = network.PathFinding( p.p, dest );
				path.cost    *= ( network.GetSubArea( dest ).Density() + 1.0 ) * gaus.Rand();

				if( path.cost < best_cost ){
					best_cost = path.cost;
					p.path = path.path;
					p.action.dest_no = field->seats.GetDataNo( p.action.DestID(), i );
				}
			}
			break;

		case GOTO_QUEUE:
			for( int i=0; i< field->queues.DataNum( p.action.DestID() ); i++ ){

				Coordinate dest = QueueLast( p.action.DestID(), i );
				PathCost path = network.PathFinding( p.p, dest );
				path.cost += queue_mng[field->queues.GetDataNo( p.action.DestID(), i )].size() * p.action.WaitTimeMean();
				path.cost *= gaus.Rand();

				if( i==0 || path.cost < best_cost ){
					best_cost = path.cost;
					p.path = path.path;
					p.action.dest_no = field->queues.GetDataNo( p.action.DestID(), i );
				}
			}
			break;
	}

	if( ReachWaypoint( p )) p.path.pop_front();
	return p;
}


// Name: Potential() -----------------------------------------------------------------------------+
// |eV@ɐlX̑xxNgXV
//------------------------------------------------------------------------------------------------+
Person Simulator::Potential( Person p, People predict ){

	// ړInւ̃|eVvZ
	Vector2D potential = DestinationPotential( p );

	// ǂ̃|eV
	potential += WallsPotential( p.p );

	// ͂̐l̃|eV
	for( int i=0; i < (int)predict.size(); i++ ) if( p.ID() != predict[i].ID() ){

		potential += PersonPotential( p.p, predict[i].p, Q_PERSON );
	}

	// ͂̃{bg̃|eV
	for( int i=0; i < (int)robots.size(); i++ ){

		potential += PersonPotential( p.p, robots[i], Q_ROBOT );
	}

	potential.Normalize();
	p.p.Normalize();
	p.p += potential;
	p.p *= 0.5 * WalkingRate( p, predict );

	return p;
}


// Name: WalkingRate() ---------------------------------------------------------------------------+
// w肵l̕sxvZ
//------------------------------------------------------------------------------------------------+
double Simulator::WalkingRate( Person p, People predict ){

	int num = 0;		// ͈͓̐l
	p.p.Normalize();	// xNg

	for( int i=0; i < (int)predict.size(); i++ ) if( p.ID() != predict[i].ID() ){

		Line potential;

		// lւ̃xNgƋvZ
		Vector2D toperson( predict[i].p.x - p.p.x, predict[i].p.y - p.p.y );
		double distance = toperson.Length();

		// ͈͊O
		if( 2.0 < distance ) continue;

		toperson.Normalize();

		// Ɛlւ̃xNg̓ = cos()
		double cos = p.p * toperson;

		// ͈͊mF
		if( 0.5 * sqrt(3.0) <= cos ) num++;
	}

	return p.WalkSpeed() - FRUIN_CONST * num / sqrt(3.0);
}


// Name: DestinationPotential() ------------------------------------------------------------------+
// ړInւ̃|eVԂ
//------------------------------------------------------------------------------------------------+
Vector2D Simulator::DestinationPotential( Person p ){

	Vector2D potential;

	if( p.path.empty() ) potential = ToDestination( p );

	// ԖړInւ̃|eV
	else{
		potential = p.path.front().Border( network.SubAreaID( p.p )).PointToLine( p.p, BORDER_SIDE );
		potential.Normalize();
		potential += NextDestPotential( p, p.path );
	}

	potential.Normalize();
	return potential;
}


// Name: ToDestination() -------------------------------------------------------------------------+
// ړInւ̃xNgԂ
//------------------------------------------------------------------------------------------------+
Vector2D Simulator::ToDestination( Person p ){

	Vector2D potential;

	switch( p.action.Type() ){
		case GOTO_EXIT:
			potential = field->exits.GetData( p.action.dest_no ).GetExit().PointToLine( p.p, BORDER_SIDE );
			return potential;

		case GOTO_POINT:
			potential.dx = field->points.GetData( p.action.dest_no ).GetX() - p.p.x;
			potential.dy = field->points.GetData( p.action.dest_no ).GetY() - p.p.y;
			return potential;

		case GOTO_SEAT:
			potential.dx = field->seats.GetData( p.action.dest_no ).GetX() - p.p.x;
			potential.dy = field->seats.GetData( p.action.dest_no ).GetY() - p.p.y;
			return potential;

		case GOTO_QUEUE:
			potential = ToQueue(p);
			return potential;
	}

	return potential;	// zO
}


// Name: NextDestPotential() ---------------------------------------------------------------------+
// Xɐ̖ړInւ̃|eVԂ
// ̌oHXVĂЂƂÕTuGAƂ̋E߂Ăւ̃xNg𓾂
// ɂ̐ċAIɋ߂
// ŏIIɂ͎́iԁjړInȂړInւ̃xNgvZďI
//------------------------------------------------------------------------------------------------+
Vector2D Simulator::NextDestPotential( Person p, Path path ){

	// oHȂiċAIj
	if( path.empty() ) return Vector2D();

	Vector2D  potential;
	SubArea nextarea = path.front();
	path.pop_front();

	if( path.empty() ) potential = ToDestination( p );
	else potential = path.front().Border( nextarea ).PointToLine( p.p, BORDER_SIDE );

	// EɓȂ΃|eVȂiċAIj
	if( field->CheckConnect( Line( p.p, potential )) == false ) return Vector2D();

	potential.Normalize();

	// ċAIɍXɐւ̃|eVZĕԂ
	return potential + NextDestPotential( p, path );
}


// Name: WallsPotential() ------------------------------------------------------------------------+
// wWł̕ǂ̃|eVԂ
//------------------------------------------------------------------------------------------------+
Vector2D Simulator::WallsPotential( Coordinate c ){

	Vector2D potential;
	double nearest_dist;

	// ł߂ǂւ̃xNg߂
	for( int w=0; w < field->WallNum(); w++ ){

		Vector2D towall = field->Wall(w).PointToLine( c );

		double towall_dist = towall.Length();

		if( w == 0 || towall_dist < nearest_dist ){

			potential    = towall;
			nearest_dist = towall_dist;
		}
	}

	// ł߂ǂւ̃xNg|eVvZ
	potential *= - Q_WALLS;
	potential /= pow( nearest_dist, 4 );

	// 臒lǂ΃|eV̓[
	if( 0 < WALL_DIST_THRESHOLD && WALL_DIST_THRESHOLD < nearest_dist ) potential *= 0.0;

	return potential;
}


// Name: PersonPotential() -----------------------------------------------------------------------+
// lpɑ΂wW̐l̑傫q̃|eVԂ
//------------------------------------------------------------------------------------------------+
Vector2D Simulator::PersonPotential( Line p, Coordinate c, double q ){

	Vector2D potential;

	// xNg
	p.Normalize();

	// lւ̃xNgƋvZ
	Line toperson( p, c );
	double distance = toperson.Length();

	// O
	if( VIEW_RANGE < distance ) return potential;

	// ԂɕǂƂ
	if( !field->CheckConnect( toperson ) ) return potential;

	toperson.Normalize();

	// Ɛlւ̃xNg̓ = cos()
	double cos = p * toperson;

	// O
	if( cos < 0.5 ) return potential;

	// |eVvZ
	potential = toperson;
	potential *= - q * cos;
	potential /= pow( distance, 2 );

	return potential;
}


// Name: CheckSeatsFree() ------------------------------------------------------------------------+
// eȂ󂢂Ă邩𒲂ׂ
//------------------------------------------------------------------------------------------------+
void Simulator::CheckSeatsFree(){

	for( int s=0; s< field->seats.DataNum(); s++ ) if( seats_mng[s].sitting = true ){

		int id = PersonNo( seats_mng[s].id );

		if( id == -1 ){
			seats_mng[s].sitting = false;
			continue;
		}

		Vector2D seat_to_person = Vector2D( people[id].p.x - field->seats.GetData(s).GetX(),
											people[id].p.y - field->seats.GetData(s).GetY()
								);

		if( SEAT_RADIUS < seat_to_person.Length() ){
			seats_mng[s].sitting = false;
			continue;
		}
	}
}


// Name: PersonNo() ------------------------------------------------------------------------------+
// ID炻̐l̓YԂ
//------------------------------------------------------------------------------------------------+
int Simulator::PersonNo( int id ){

	for( int p=0; p<(int)people.size(); p++ ) if( people[p].ID() == id ) return p;

	return -1;	// error;
}


// Name: QueueLast() -----------------------------------------------------------------------------+
// ws̍ŌWԂ
//------------------------------------------------------------------------------------------------+
Coordinate Simulator::QueueLast( int id, int no ){

	int q_no = field->queues.GetDataNo( id, no );

	// lȂΐ擪
	if( queue_mng[q_no].empty() ) return field->queues.GetData( q_no ).Data();

	Coordinate last;
	last.x = people[PersonNo(queue_mng[q_no].back())].p.x + field->queues.GetData( q_no ).Data().dx;
	last.y = people[PersonNo(queue_mng[q_no].back())].p.y + field->queues.GetData( q_no ).Data().dy;

	return last;
}


// Name: QueuePosition() -------------------------------------------------------------------------+
// ړIn̍s̖ڎwׂWԂ
//------------------------------------------------------------------------------------------------+
Coordinate Simulator::QueuePosition( Person p ){

	Coordinate to_q;

	if( p.action.Type() != GOTO_QUEUE ) return to_q;

	// ̐擪ڎwƂ
	if( queue_mng[p.action.dest_no].empty() || queue_mng[p.action.dest_no].front() == p.ID() ){
		to_q.x = field->queues.GetData( p.action.dest_no ).GetX();
		to_q.y = field->queues.GetData( p.action.dest_no ).GetY();
		return to_q;
	}

	// ̑OiɕłƂj܂͍ŌiłȂƂj̐l𒊏o
	std::list<int> list = queue_mng[p.action.dest_no];
	int vor_id = list.front();
	list.pop_front();

	for(; !list.empty(); list.pop_front() ){
		if( list.front() == p.ID() ) break;
		else vor_id = list.front();
	}

	// ڎwׂWvZĕԂ
	to_q.x = people[PersonNo(vor_id)].p.x + field->queues.GetData( p.action.dest_no ).Data().dx;
	to_q.y = people[PersonNo(vor_id)].p.y + field->queues.GetData( p.action.dest_no ).Data().dy;

	return to_q;
}


// Name: ToQueue() -------------------------------------------------------------------------------+
// ړIn̍sւ̃xNgԂ
//------------------------------------------------------------------------------------------------+
Line Simulator::ToQueue( Person p ){

	return Line( p.p, QueuePosition( p ) );
}


// Name: OutQueue() ------------------------------------------------------------------------------+
// w肵l񂩂
//------------------------------------------------------------------------------------------------+
void Simulator::OutQueue( Person p ){

	std::list<int> update;

	for( std::list<int> list = queue_mng[p.action.dest_no]; !list.empty(); list.pop_front() ){
		if( list.front() != p.ID() ) update.push_back( list.front() );
	}

	queue_mng[p.action.dest_no] = update;
}

