//------------------------------------------------------------------------------------------------+
// Filename: Triangulation.cpp
// Content : Generate triangules and convert to convex cells
//------------------------------------------------------------------------------------------------+

#include <cassert>
#include <iostream>

#include "Triangulation.h"
#include "cv.h"
#include "highgui.h"

#define ONEMICROMETER	0.000001	// 1m͌덷ƂĈ߂̃}N


// Name: PushNewPoint() --------------------------------------------------------------------------+
// ̒_܂i[ĂȂΊi[trueԂ
//------------------------------------------------------------------------------------------------+
Tessellator::PointID Tessellator::PushNewPoint( Vertex vertex, PointType type ){

	for( int p=0; p<(int)points.size(); p++ ) if( vertex == points[p] || Segment( vertex, points[p] ).Length() < ONEMICROMETER ) return (PointID) p;

	points.push_back( Point( vertex, type ) );
	DivideLink( (PointID)(points.size()-1) );

	return (PointID)( (int)points.size() - 1 );
}


// Name: PushNewLink() ---------------------------------------------------------------------------+
// ̃N܂i[ĂȂΊi[
// WȂłΓNƂ݂Ȃ
//------------------------------------------------------------------------------------------------+
Tessellator::LinkID Tessellator::PushNewLink( Link link ){

	for( int l=0; l<(int)links.size(); l++ ){
		if( link.point_id[0] == links[l].point_id[0] && link.point_id[1] == links[l].point_id[1] ) return (LinkID)l;
		if( link.point_id[0] == links[l].point_id[1] && link.point_id[1] == links[l].point_id[0] ) return (LinkID)l;
	}

	links.push_back( link );

	// ӂƂ̌_Ίi[
	Segment seg( points[link.point_id[0]], points[link.point_id[1]] );
	for( int l=0; l<(int)links.size(); l++ ) if( LinkSegment(l).KreuzungUrteilX( seg ) ){

		PointType p_type = ( link.type == BOTTLENECK )? GENERATED_POINT: ORIGINAL_POINT;
		PushNewPoint( LinkSegment(l).Kreuzung( seg ), p_type );
	}

	return (LinkID)( links.size() - 1 );
}


// Name: PushNewLink() ---------------------------------------------------------------------------+
// ̐NƂĂ܂i[ĂȂΊi[
// WȂłΓNƂ݂Ȃ
//------------------------------------------------------------------------------------------------+
Tessellator::LinkID Tessellator::PushNewLink( Segment seg, LinkType type ){

	PointType p_type = ( type == BOTTLENECK )? GENERATED_POINT: ORIGINAL_POINT;

	PointID p_id1 = PushNewPoint( seg, p_type );
	PointID p_id2 = PushNewPoint( seg.Inverse(), p_type );
	Link link( p_id1, p_id2 );
	link.type = type;

	return PushNewLink( link );
}


// Name: PushNewLink() ---------------------------------------------------------------------------+
// ̍WȂNƂĂ܂i[ĂȂΊi[
// WȂłΓNƂ݂Ȃ
//------------------------------------------------------------------------------------------------+
Tessellator::LinkID Tessellator::PushNewLink( Coordinate p1, Coordinate p2, LinkType type ){

	PointType p_type = ( type == BOTTLENECK )? GENERATED_POINT: ORIGINAL_POINT;

	PointID p_id1 = PushNewPoint( p1, p_type );
	PointID p_id2 = PushNewPoint( p2, p_type );
	Link link( p_id1, p_id2 );
	link.type = type;

	return PushNewLink( link );
}


// Name: DivideLink() ----------------------------------------------------------------------------+
// wID̒_ӏɂΕӂ𕪊
//------------------------------------------------------------------------------------------------+
bool Tessellator::DivideLink( PointID id ){

	bool check = false;
	int links_size = (int)links.size();

	for( int l=0; l<links_size; l++ ) if( LinkSegment(l).PointToLine(points[id]).Length() < ONEMICROMETER ){

		check = true;

		Link link0( links[l].point_id[0], id, links[l].type );
		Link link1( links[l].point_id[1], id, links[l].type );

		links[l] = link0;
		links.push_back( link1 );
	}

	return check;
}


// Name: IsNeighbor() ----------------------------------------------------------------------------+
// IDŎw肵ʂƗאڂĂtrueԂ
//------------------------------------------------------------------------------------------------+
bool Tessellator::IsNeighbor( FaceID face1, FaceID face2 ){

	if( face1 < 0 || (int)faces.size() <= face1 ) return false;
	if( face2 < 0 || (int)faces.size() <= face2 ) return false;

	// ʂ̕ӂ邩TΗǂ
	for( int i=0; i<(int)faces[face1].links.size(); i++ ){
		for( int j=0; j<(int)faces[face2].links.size(); j++ ){

			if( faces[face1].links[i] == faces[face2].links[j] ) return true;
		}
	}

	return false;
}


// Name: ExistPoint() ----------------------------------------------------------------------------+
// wW_f[^ɑ݂trueԂ
//------------------------------------------------------------------------------------------------+
bool Tessellator::ExistPoint( Vertex vertex ){

	for( int p=0; p<(int)points.size(); p++ ) if( vertex == points[p] || Segment( vertex, points[p] ).Length() < ONEMICROMETER ) return true;
	return false;
}


// Name: ExistLink() -----------------------------------------------------------------------------+
// w_ԃNo^ς݂łtrueԂ
//------------------------------------------------------------------------------------------------+
bool Tessellator::ExistLink( PointID p1, PointID p2 ){

	for( int l=0; l<(int)links.size(); l++ ){
		if( p1 == links[l].point_id[0] && p2 == links[l].point_id[1] ) return true;
		if( p1 == links[l].point_id[1] && p2 == links[l].point_id[0] ) return true;
	}

	return false;
}


// Name: FirstTriangulation() --------------------------------------------------------------------+
// i[Ă_ƃN̏񂩂CDTꂩ\z
//------------------------------------------------------------------------------------------------+
void Tessellator::FirstTriangulation(){

	CDT triangulation;

	for( int i=0; i<(int)links.size(); i++ ){

		triangulation.insert_constraint(
			CDT::Point( points[links[i].point_id[0]].x, points[links[i].point_id[0]].y ),
			CDT::Point( points[links[i].point_id[1]].x, points[links[i].point_id[1]].y )
			);
	}

	cdt = triangulation;
	BuildTriangles();
//	for( int p=0; p<(int)points.size(); p++ ) cout << "Point[" << p << "]" << " = ( " << points[p].x << ", " << points[p].y << " );\n";
}


// Name: BuildTriangles() ------------------------------------------------------------------------+
// CDŤʂOp`f[^n𐶐
//------------------------------------------------------------------------------------------------+
void Tessellator::BuildTriangles(){

	assert(cdt.is_valid());

	vector<Face> triangles;

	for( CDT::All_faces_iterator afi = cdt.all_faces_begin(); afi != cdt.all_faces_end(); afi++ ) if( afi->is_valid() ){

		Vertexes vert;

		bool check = true;

		for( int i=0; i<3; i++ ){
			vert.push_back( Coordinate( afi->vertex(i)->point().x(), afi->vertex(i)->point().y()));
			if( !ExistPoint( vert[i] )) check = false;
		}
		if( !check ) continue;

		Face face( vert );

		if( !IsInsideEmpty( face ) ) continue;

		for( int i=0; i<3; i++ ){

			LinkID link_id = PushNewLink( vert[i], vert[(i+1)%3], FREE_EDGE );

			if( links[link_id].face_id[0] == NO_LINK_FACE ){

				 links[link_id].face_id[0] = (FaceID)triangles.size();
			}
			else links[link_id].face_id[1] = (FaceID)triangles.size();

			face.links.push_back( link_id );
		}

		triangles.push_back( face );

//		SavePointLinkImage( 40.0 );
	}

	faces = triangles;
}


// Name: DetectBottlenecks() ---------------------------------------------------------------------+
// {glbNoē_ƃNɉ
//------------------------------------------------------------------------------------------------+
void Tessellator::DetectBottlenecks( Segments constrained ){

	Segments bottlenecks;

	// _Ɛ̍ŒZo
	for( int p=0; p<(int)points.size(); p++ ) for( int c=0; c<(int)constrained.size(); c++ ){

		Segment bottleneck = constrained[c].PointToLine( points[p] );

		if( bottleneck.Length() == 0.0 ) continue;

		// sǂmF
		if( 0.0000001 < Abs( bottleneck * constrained[c] ) ) continue;

		// ̒_ɂȂmF
		bool closs = false;
		for( int q=0; q<(int)points.size(); q++ ){
			if( p != q && bottleneck.PointToLine( points[q] ).Length() < ONEMICROMETER  ){

				if( points[q] == (Coordinate)bottleneck.Inverse() ) continue; 
				closs = true;
				break;
			}
		}
		if( closs ) continue;

		// o{glbNƓNłɓo^ς݂łȂmF
		if( ExistPoint( bottleneck.Inverse() ) && ExistLink( p, PushNewPoint( bottleneck.Inverse(), GENERATED_POINT ))) continue;

		// ̐ƌȂmF
		for( int d=0; d<(int)constrained.size(); d++ ) if( c != d ){

			if( bottleneck.KreuzungUrteilX( constrained[d] ) ){

				closs = true;
				break;
			}
		}
		if( !closs ) bottlenecks.push_back( bottleneck );
	}

	// {glbN΁A̒ŒẐ̂I
	Segments select;
	for( int b=0; b<(int)bottlenecks.size(); b++ ){

		bool check = true;

		for( int c=0; c<(int)bottlenecks.size(); c++ ) if( b != c ){

			if( bottlenecks[b].KreuzungUrteilX( bottlenecks[c] ) && bottlenecks[c].Length() <= bottlenecks[b].Length() ){

				check = false;
				break;
			}
		}
		if( check ) select.push_back( bottlenecks[b] );
	}

	// _Ɛ̃f[^Ƀ{glbN
	for( int b=0; b<(int)select.size(); b++ ) PushNewLink( select[b], BOTTLENECK );
}


// Name: IsInsideEmpty() -------------------------------------------------------------------------+
// ʂ̓Ƀ|CgȂtrueԂ
//------------------------------------------------------------------------------------------------+
bool Tessellator::IsInsideEmpty( Face face ){

	// _ɂ邩̔
	for( int p=0; p<(int)points.size(); p++ ){

		bool check_same_point = false;
		for( int v=0; v<(int)face.Vertex().size(); v++ ){
			if( Segment( face.Vertex()[v], points[p] ).Length() < ONEMICROMETER ){

				check_same_point = true;
				break;
			}
		}
		if( !check_same_point && face.InSide( points[p] ) ) return false;
	}

	// Ƃ̌
	Segments edges;

	for( int v=0; v<(int)face.Vertex().size(); v++ ){

		edges.push_back( Segment( Segment( face.Vertex()[v], face.Vertex()[(v+1)%face.Vertex().size()] )));
	}

	for( int j=0; j<(int)links.size(); j++ ){

		Segment seg( points[links[j].point_id[0]], points[links[j].point_id[1]] );

		for( int i=0; i<(int)edges.size(); i++ ){

			if( edges[i].KreuzungUrteilX( seg ) ) return false;
		}
	}
	return true;
}


// Name: SortLinks() -----------------------------------------------------------------------------+
// N𒷂̍~Ń\[gIĎnԂ
//------------------------------------------------------------------------------------------------+
vector<Tessellator::LinkID> Tessellator::SortLinks(){

	vector<LinkID> sorted_id;
	vector<bool> sorted;
	sorted.resize( links.size() );

	for( int i=0; i<(int)sorted.size(); i++ ) sorted[i] = false;

	while( sorted_id.size() != (int)links.size() ){

		LinkID longst_id = 0;
		double longst_length = 0.0;
		for( int i=1; i<(int)links.size(); i++ ){
			if( sorted[i] ) continue;
			if( sorted[longst_id] ) longst_id = i;
			else{
				double length = Segment( points[links[i].point_id[0]], points[links[i].point_id[1]] ).Length();
				if( longst_length < length ){

					longst_length = length;
					longst_id = (LinkID)i;
				}
			}
		}

		sorted[longst_id] = true;
		sorted_id.push_back( longst_id );
	}

	return sorted_id;
}


// Name: Optimization() --------------------------------------------------------------------------+
// RŐڂ鑽p`ʌ^͈̔͂œf[^
//------------------------------------------------------------------------------------------------+
void Tessellator::Optimization(){

	cout << "\nTriangules are merged into convex polygons.\n";

	vector<LinkID> sorted_id = SortLinks();

	// ёւRԂɏ
	for( int i=0; i<(int)sorted_id.size(); i++ ){

		// Rǂ
		if( links[sorted_id[i]].type != FREE_EDGE ) continue;

		FaceID face_id[2];

		// RŐڂQʂ𒊏o
		for( int j=0; j<2; j++ ){

			face_id[j] = links[sorted_id[i]].face_id[j];

			// ؍\̃[głʂID擾
			while( face_id[j] != NO_LINK_FACE && faces[face_id[j]].parent_id != NO_PARENT )
				face_id[j] = faces[face_id[j]].parent_id;
		}

		// RQʂƐڂĂȂӁiԂ̒[jꍇł͂Ȃ
		if( face_id[0] == NO_LINK_FACE || face_id[1] == NO_LINK_FACE ) continue;

		Face compound( faces[face_id[0]] );

		if( !compound.ConvexIntegration( faces[face_id[1]] )) continue;

		vector<LinkID> temp_link_id;

		for( int j=0; j<2; j++ ){

			// eƎq̓o^
			compound.childlen[j] = face_id[j];
			faces[face_id[j]].parent_id = (int)faces.size();

			// ̖ʂĂ鎩RiӁj̃f[^i[邽
			for( int e=0; e<(int)faces[face_id[j]].links.size(); e++ ){

				if( faces[face_id[j]].links[e] == sorted_id[i] ) continue;
				temp_link_id.push_back( faces[face_id[j]].links[e] );
			}
		}
		compound.links = temp_link_id;

		faces.push_back( compound );
	}
}


// Name: SavePointLinkImage() --------------------------------------------------------------------+
// _Ɛō\zꂽf[^\摜ɏo͂
//------------------------------------------------------------------------------------------------+
void Tessellator::SavePointLinkImage( double scale ){

	double width = 0, height = 0;
	for( int p=0; p<(int)points.size(); p++ ){
		if( width  < points[p].x * scale ) width  = points[p].x * scale;
		if( height < points[p].y * scale ) height = points[p].y * scale;
	}

	IplImage *img = cvCreateImage( cvSize( (int)width+1, (int)height+1 ), IPL_DEPTH_8U, 3 );

	cvSet( img, cvScalarAll(255), NULL );

	for( int i=0; i<(int)links.size(); i++ ){
		CvPoint from = { (int)( points[links[i].point_id[0]].x * scale ), (int)( height - points[links[i].point_id[0]].y * scale ) };
		CvPoint to   = { (int)( points[links[i].point_id[1]].x * scale ), (int)( height - points[links[i].point_id[1]].y * scale ) };
		cvLine( img, from, to, CV_RGB( (links[i].type==BOTTLENECK)?255:0, 0, (links[i].type==FREE_EDGE)?255:0 ), (links[i].type==CONSTRAINED)?3:1, 8, 0 );
	}

	for( int p=0; p<(int)points.size(); p++ ){
		cvCircle( img, cvPoint( (int)( points[p].x * scale ), (int)( height - points[p].y * scale ) ), 6, CV_RGB( (points[p].type==GENERATED_POINT)?255:0, (points[p].type==ORIGINAL_POINT)?255:0, 0 ), 2, 8, 0 );
//		CvFont dfont;
//		cvInitFont( &dfont, CV_FONT_HERSHEY_SIMPLEX , 0.5, 0.5, 0.0, 1, CV_AA);
//		char str[10];
//		sprintf( str, "%d",p );
//		cvPutText( img, str, cvPoint((int)(points[p].x * scale), (int)(height - points[p].y * scale)), &dfont, CV_RGB(0, 0, 0));
	}

	char filename[20];
	sprintf( filename, "Images/%04d.png", image_count++ );
	cvSaveImage( filename, img );
	cvReleaseImage( &img );
}


// Name: SavePolygonsImage() ---------------------------------------------------------------------+
// p`f[^摜ɏo͂
//------------------------------------------------------------------------------------------------+
void Tessellator::SavePolygonsImage( double scale ){

	double width = 0, height = 0;
	for( int p=0; p<(int)points.size(); p++ ){
		if( width  < points[p].x * scale ) width  = points[p].x * scale;
		if( height < points[p].y * scale ) height = points[p].y * scale;
	}

	IplImage *img = cvCreateImage( cvSize( (int)width+1, (int)height+1 ), IPL_DEPTH_8U, 3 );

	cvSet( img, cvScalarAll(255), NULL );

	for( int i=0; i<(int)faces.size(); i++ ) if( faces[i].parent_id == NO_PARENT ) {

		for( int j=0; j<(int)faces[i].Vertex().size(); j++ ){

			CvPoint from = { (int)( faces[i].Vertex()[j].x * scale ), (int)( height - faces[i].Vertex()[j].y * scale ) };
			CvPoint to   = { (int)( faces[i].Vertex()[(j+1)%faces[i].Vertex().size()].x * scale ), (int)( height - faces[i].Vertex()[(j+1)%faces[i].Vertex().size()].y * scale ) };
			cvLine( img, from, to, CV_RGB( 0, 0, 255 ), 1, 8, 0 );
		}
	}

	for( int i=0; i<(int)links.size(); i++ ) if( links[i].type == CONSTRAINED ){
		CvPoint from = { (int)( points[links[i].point_id[0]].x * scale ), (int)( height - points[links[i].point_id[0]].y * scale ) };
		CvPoint to   = { (int)( points[links[i].point_id[1]].x * scale ), (int)( height - points[links[i].point_id[1]].y * scale ) };
		cvLine( img, from, to, CV_RGB( 0, 0, 0 ), 3, 8, 0 );
	}

	cvSaveImage( "Images/Polygons.png", img );
	cvReleaseImage(&img);
}


// Name: GenerateSubAreas() ----------------------------------------------------------------------+
// ʌ^p`TuGAɕϊ
//------------------------------------------------------------------------------------------------+
SubAreas Tessellator::GenerateSubAreas( double max_link_length ){

	SubAreas subareas;

	// TuGAɓo^
	for( int p=0; p<(int)faces.size(); p++ ){

		if( faces[p].parent_id != NO_PARENT ) continue;

		subareas.push_back( SubArea( p, faces[p].Vertex()));
	}

	// No^
	for( int e=0; e<(int)links.size(); e++ ){

		if( links[e].type == CONSTRAINED ) continue;
		if( links[e].face_id[0] == NO_LINK_FACE || links[e].face_id[1] == NO_LINK_FACE ) continue; 
		if( 0.0 < max_link_length && LinkSegment(e).Length() < max_link_length ) continue;

		int face_id[2];

		for( int i=0; i<2; i++ ){

			face_id[i] = links[e].face_id[i];

			while( faces[face_id[i]].parent_id != NO_PARENT ) face_id[i] = faces[face_id[i]].parent_id;
		}

		if( face_id[0] == face_id[1] ) continue;

		int subarea_id[2];

		for( int i=0; i<2; i++ ) for( int j=0; j<(int)subareas.size(); j++){

			if( subareas[j].ID() == face_id[i] ){
				subarea_id[i] = j;
				break;
			}
		}

		for( int i=0; i<2; i++ ){

			subareas[subarea_id[i]].AddLink( face_id[!i], LinkSegment(e) );
		}
	}

	return subareas;
}


