//------------------------------------------------------------------------------------------------+
// Filename: Triangulation.h
// Content : Generate triangules and merge to convex cells
//------------------------------------------------------------------------------------------------+
#ifndef __TRIANGULATION
#define __TRIANGULATION

#include "applib.h"
#include "Network.h"

#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Constrained_Delaunay_triangulation_2.h>


typedef CGAL::Exact_predicates_inexact_constructions_kernel K;

typedef CGAL::Triangulation_vertex_base_2<K>                     Vb;
typedef CGAL::Constrained_triangulation_face_base_2<K>           Fb;
typedef CGAL::Triangulation_data_structure_2<Vb,Fb>              TDS;
typedef CGAL::Exact_predicates_tag                               Itag;
typedef CGAL::Constrained_Delaunay_triangulation_2<K, TDS, Itag> CDT;

using namespace std;


//------------------------------------------------------------------------------------------------+
// }N
//------------------------------------------------------------------------------------------------+
#define NO_PARENT		-1			// for parent_id of GeometricData::Face
#define NO_CHILD		-1			// for base_id of GeometricData::Face
#define NO_LINK_FACE	-1			// for face_id of GeometricData::Link


//------------------------------------------------------------------------------------------------+
// ʌ^p`NX
//------------------------------------------------------------------------------------------------+
class Tessellator{

	typedef int PointID;
	typedef int LinkID;
	typedef int FaceID;

	enum PointType{ ORIGINAL_POINT, GENERATED_POINT };
	enum LinkType{ CONSTRAINED, FREE_EDGE, BOTTLENECK };

	// _
	class Point : public Vertex{
	public:
		PointType type;
		Point( Vertex vertex, PointType type ){
			this->x = vertex.x;
			this->y = vertex.y;
			this->type = type;
		};
	};

	// _Ɠ_ԃN
	class Link{
	public:
		LinkType type;
		PointID point_id[2];
		FaceID face_id[2];

		Link( PointID point1, PointID point2, LinkType type=CONSTRAINED ){
			this->type = type;
			point_id[0] = point1;
			point_id[1] = point2;
			face_id[0] = face_id[1] = NO_LINK_FACE;
		};
	};

	// 
	class Face : public Vieleck{
	public:
		FaceID parent_id;
		FaceID childlen[2];
		vector<LinkID> links;

		Face( Vertexes vert ):Vieleck( vert ){
			parent_id = NO_PARENT;
			childlen[0] = childlen[1] = NO_CHILD;
		}
	};

	vector<Point> points;
	vector<Link> links;
	vector<Face> faces;

	CDT cdt;

	int image_count;

	// _ƃN̊ǗɊւ֐Q
	PointID PushNewPoint( Vertex vertex, PointType type );
	LinkID PushNewLink( Link link );
	LinkID PushNewLink( Segment seg, LinkType type );
	LinkID PushNewLink( Coordinate p1, Coordinate p2, LinkType type );
	bool DivideLink( PointID id );
	bool IsNeighbor( FaceID face1, FaceID face2 );
	bool ExistPoint( Vertex vertex );
	bool ExistLink( PointID p1, PointID p2 );
	
	Segment LinkSegment( LinkID link ){
		return Segment( points[links[link].point_id[0]], points[links[link].point_id[1]] );
	}

	// Op`Ɋւ֐Q
	void FirstTriangulation();
	void BuildTriangles();
	void DetectBottlenecks( Segments constrained );
	bool IsInsideEmpty( Face face );

	// p`œKɊւ֐Q
	vector<LinkID> SortLinks();
	void Optimization();

public:
	Tessellator( Segments constrained ){
		image_count = 0;
		for( int i=0; i<(int)constrained.size(); i++ ) PushNewLink( constrained[i], CONSTRAINED );
		SavePointLinkImage( 40.0 );
		DetectBottlenecks( constrained );
		SavePointLinkImage( 40.0 );
		FirstTriangulation();
		SavePointLinkImage( 40.0 );
		Optimization();
		SavePolygonsImage( 40.0 );
	}
	void SavePointLinkImage( double scale );
	void SavePolygonsImage( double scale );
	SubAreas GenerateSubAreas( double max_link_length=0.0 );
};


#endif	//__TRIANGULATION
