//------------------------------------------------------------------------------------------------+
// Filename: Window.cpp
// Content : Define of Window class's functions about create and manage a window using OpenCV.
//------------------------------------------------------------------------------------------------+

#include "Window.h"
#include <iostream>


/***************************************** WindowNX ******************************************/

// Name: Window() --------------------------------------------------------------------------------+
// RXgN^
//------------------------------------------------------------------------------------------------+
Window::Window( const char window_name[], int width, int height ){

	this->width  = width;
	this->height = height;

	for( int i=0; window_name[i]!='\0'; i++ ){

		this->window_name[i] = window_name[i];
		this->window_name[i+1] = '\0';
	}

	// EBhE̐
	cvNamedWindow( window_name, CV_WINDOW_AUTOSIZE );

	// 摜𐶐
	img = cvCreateImage( cvSize( width, height ), IPL_DEPTH_8U, 3 );

	// 摜𔒂ŏ
	cvSet( img, cvScalarAll(255), NULL );
}


// Name: Window() --------------------------------------------------------------------------------+
// Rs[RXgN^
//------------------------------------------------------------------------------------------------+
Window::Window(const Window& obj){

	std::cerr << "Don't copy!\n";
	throw;
}


// Name: ~Window() -------------------------------------------------------------------------------+
// fXgN^
//------------------------------------------------------------------------------------------------+
Window::~Window(){

	// 摜J
	cvReleaseImage(&img);

	// EBhEj
	cvDestroyWindow( window_name );
}


// Name: Show() ----------------------------------------------------------------------------------+
// 摜`悷
//------------------------------------------------------------------------------------------------+
void Window::Show(){

	// 摜\
	cvShowImage( window_name, img );
	cvWaitKey(1);
}


// Name: Reset() ---------------------------------------------------------------------------------+
// 摜𔒂ŏ
//------------------------------------------------------------------------------------------------+
void Window::Reset(){

	cvSet( img, cvScalarAll(255), NULL );
}



/************************************ SimulationWindowNX *************************************/

// Name: SimulationWindow() ----------------------------------------------------------------------+
// RXgN^
//------------------------------------------------------------------------------------------------+
SimulationWindow::SimulationWindow(const char window_name[], int width, int height, double scale)
:Window( window_name, width+1, height+1)
{
	this->scale = scale;
}


// Name: DrowPerson() ----------------------------------------------------------------------------+
// li~j`
//------------------------------------------------------------------------------------------------+
void SimulationWindow::DrowPerson( Line l, int r, int g, int b, double radius ){

	// ~ƌ\`
	l.Normalize();
	l *= 2.0 * radius;
	DrowLine( l, r, g, b, 1 );
	DrowPerson( (Coordinate)l, r, g, b, radius );
}


// Name: DrowPerson() ----------------------------------------------------------------------------+
// li~j`
//------------------------------------------------------------------------------------------------+
void SimulationWindow::DrowPerson( Coordinate p, int r, int g, int b, double radius ){

	// ~`ēhԂ
	CvPoint center = { (int)( p.x * scale ), height - (int)( p.y * scale ) };
	cvCircle( img, center, (int)( radius * scale ), CV_RGB( r, g, b ), CV_FILLED );
}


// Name: DrowWall() ------------------------------------------------------------------------------+
// ǁij`
//------------------------------------------------------------------------------------------------+
void SimulationWindow::DrowWall( Line wall ){

	CvPoint from = { (int)( wall.x * scale ), height - (int)( wall.y * scale ) - 1 };
	CvPoint to   = { (int)( wall.Inverse().x * scale ), height - (int)( wall.Inverse().y * scale ) - 1 };
	cvLine( img, from, to, CV_RGB( 0, 0, 0 ), 1, 8, 0 );
}


// Name: DrowLine() ------------------------------------------------------------------------------+
// `
//------------------------------------------------------------------------------------------------+
void SimulationWindow::DrowLine( Line line, int r, int g, int b, int thickness ){

	CvPoint from = { (int)( line.x * scale ), height - (int)( line.y * scale ) - 1 };
	CvPoint to   = { (int)( line.Inverse().x * scale ), height - (int)( line.Inverse().y * scale ) - 1 };
	cvLine( img, from, to, CV_RGB( r, g, b ), thickness, 8, 0 );
}


// Name: Print() ---------------------------------------------------------------------------------+
// `悷
//------------------------------------------------------------------------------------------------+
void SimulationWindow::Print( char *str, Coordinate p, int r, int g, int b ){

	CvFont dfont;
	cvInitFont( &dfont, CV_FONT_HERSHEY_SIMPLEX , 0.5, 0.5, 0.0, 1, CV_AA);
	cvPutText( img, str, cvPoint((int)(p.x * scale), height - (int)(p.y * scale)), &dfont, CV_RGB(r, g, b));
}


// Name: Print() ---------------------------------------------------------------------------------+
// `悷
//------------------------------------------------------------------------------------------------+
void SimulationWindow::Print( char *str, int x, int y, int r, int g, int b ){

	CvFont dfont;
	cvInitFont( &dfont, CV_FONT_HERSHEY_SIMPLEX , 1.0, 1.0, 0.0, 1, CV_AA);
	cvPutText( img, str, cvPoint(x, y), &dfont, CV_RGB(r, g, b));
}


// Name: SaveImage() -----------------------------------------------------------------------------+
// 摜ۑ
//------------------------------------------------------------------------------------------------+
void SimulationWindow::SaveImage( const char *savename ){

	cvSaveImage( savename, img );
}



/************************************* GridMapWindow NX **************************************/

// Name: GridMapWindow() -------------------------------------------------------------------------+
// RXgN^
//------------------------------------------------------------------------------------------------+
GridMapWindow::GridMapWindow(const char window_name[], int width, int height):Window( window_name, width, height){}


// Name: Put() -----------------------------------------------------------------------------------+
// wfwFɂ
//------------------------------------------------------------------------------------------------+
void GridMapWindow::Put(int x, int y, int r, int g, int b){

	if( x < 0 || y < 0 || width <= x || height <= y ) return;

	Ired(   img, x, y ) = r;
	Igreen( img, x, y ) = g;
	Iblue(  img, x, y ) = b;
}

