//------------------------------------------------------------------------------------------------+
// Filename: Window.h
// Content : Define of Window class that create and manage a window using OpenCV.
//------------------------------------------------------------------------------------------------+
#ifndef __WINDOW
#define __WINDOW

#include "cv.h"
#include "highgui.h"
#include "applib.h"


//------------------------------------------------------------------------------------------------+
// Define of macro
//------------------------------------------------------------------------------------------------+
#define MAX_WINDOW_NAME	255
#define PERSON_RADIUS	0.225

#define Iblue(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[3*(x)]		// 24rbgJ[A
#define Igreen(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[3*(x)+1]	// 24rbgJ[A
#define Ired(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[3*(x)+2]	// 24rbgJ[A



//------------------------------------------------------------------------------------------------+
// Define of Window class
//------------------------------------------------------------------------------------------------+
class Window{

protected:
	int width, height;
	IplImage *img;
	char window_name[MAX_WINDOW_NAME];

public:
	Window( const char window_name[], int width, int height );
	Window( const Window& obj );
	virtual ~Window();
	void Show();
	void Reset();

	int Width(){ return width; }
	int Height(){ return height; } 
};


//------------------------------------------------------------------------------------------------+
// Define of Simulation Window class
//------------------------------------------------------------------------------------------------+
class SimulationWindow :public Window{

	double scale;

public:
	SimulationWindow( const char window_name[], int width, int height, double scale );
	void DrowPerson( Line l, int r=0, int g=255, int b=0, double radius=PERSON_RADIUS );
	void DrowPerson( Coordinate p, int r=0, int g=255, int b=0, double radius=PERSON_RADIUS );
	void DrowWall( Line wall );
	void DrowLine( Line line, int r=0, int g=0, int b=0, int thickness=1 );
	void Print( char *str, Coordinate p, int r=0, int g=0, int b=0 );
	void Print( char *str, int x=0, int y=0, int r=0, int g=0, int b=0 );
	void SaveImage( const char *savename );
};


//------------------------------------------------------------------------------------------------+
// Define of pixel editable Window class
//------------------------------------------------------------------------------------------------+
class GridMapWindow :public Window{

public:
	GridMapWindow( const char window_name[], int width, int height );
	void Put( int x, int y, int r, int g, int b );
};


#endif // __WINDOW
